"""Stale Packages CLI."""

from invoke.context import Context

from ..check_for_stale_packages import check_for_stale_packages as cfsp
from ..cli import task
from ..invoke_helpers import run
from .executable_utils import python_dir


@task(
    default=True,
    help={
        'stale_months': 'Cutoff in months for when a package may be stale enough to be a risk',
    },
)
def check_for_stale_packages(ctx: Context, *, stale_months: int = 48) -> None:
    """Identify stale dependencies."""
    cfsp(stale_months=stale_months)
    run(ctx, f'{python_dir()}/pip-check --cmd="poetry run pip" --hide-unchanged')



"""nox-poetry configuration file.

[Useful snippets from docs](https://nox.thea.codes/en/stable/usage.html)

```sh
poetry run nox -l
poetry run nox --list-sessions

poetry run nox -s build_check-3.8 build_dist-3.8 tests-3.8
poetry run nox --session tests-3.11

poetry run nox --python 3.8

poetry run nox -k "not build_check and not build_dist"
```

Useful nox snippets

```python3
# Example conditionally skipping a session
if not session.interactive:
    session.skip('Cannot run detect-secrets audit in non-interactive shell')

# Install pinned version
session.install('detect-secrets==1.0.3')

# Example capturing STDOUT into a file (could do the same for stderr)
path_stdout = Path('.stdout.txt').resolve()
with open(path_stdout, 'w') as out:
    session.run(*shlex.split('echo Hello World!'), stdout=out)
```

"""

import shlex
from functools import lru_cache
from pathlib import Path
from urllib.parse import urlparse
from urllib.request import url2pathname

from beartype import beartype
from beartype.typing import Dict, List, Union, cast
from corallium import file_helpers  # Required for mocking read_pyproject
from corallium.file_helpers import get_tool_versions, if_found_unlink, read_package_name
from corallium.log import logger
from nox import Session as NoxSession
from nox_poetry import session as nox_poetry_session
from nox_poetry.poetry import DistributionFormat
from nox_poetry.sessions import Session as NPSession

if read_package_name() == 'corallium':
    # 'poetry export' will fail on circular dependencies, so use no instead of nox-poetry
    from nox import session as nox_session
else:
    from nox_poetry import session as nox_session  # type: ignore[no-redef]


@lru_cache(maxsize=1)
@beartype
def _get_pythons() -> List[str]:
    """Get python versions from the `.tool-versions` file."""
    return [*{str(ver) for ver in get_tool_versions()['python']}]


@beartype
def _retrieve_keys(source: Dict, keys: List[str]) -> Dict:  # type: ignore[type-arg]
    """Retrieve nested dictionary keys unless not found."""
    result = source
    for key in keys:
        if not (result := result.get(key)):  # type: ignore[assignment]
            return {}
    return result


@beartype
def _get_poetry_dev_dependencies() -> Dict[str, Dict]:  # type: ignore[type-arg]
    """Return a dictionary of all dev-dependencies from the 'pyproject.toml'."""
    poetry_config = file_helpers.read_pyproject()['tool']['poetry']

    @beartype
    def normalize_dep(value: Union[str, Dict]) -> Dict:  # type: ignore[type-arg]
        return {'version': value} if isinstance(value, str) else value

    return {
        key: normalize_dep(value) for key, value in {
            **_retrieve_keys(poetry_config, ['dev', 'dependencies']),
            **_retrieve_keys(poetry_config, ['group', 'dev', 'dependencies']),
        }.items()
    }


@lru_cache(maxsize=1)
@beartype
def _installable_dev_dependencies() -> List[str]:
    """list of dependencies from pyproject, excluding calcipy.

    Returns:
        List[str]: `['Cerberus=>1.3.4', 'freezegun']`

    """
    @beartype
    def to_package(key: str, value: Dict) -> str:  # type: ignore[type-arg]
        extras = value.get('extras', [])
        return f'{key}[{",".join(extras)}]' if extras else key

    @beartype
    def to_constraint(value: Dict) -> str:  # type: ignore[type-arg]
        return str(value['version']).replace('^', '==')

    return [
        f'{to_package(key, value)}{to_constraint(value)}'
        for key, value in _get_poetry_dev_dependencies().items()
        if key != 'calcipy'
    ]


@beartype
def _install_local(session: Union[NoxSession, NPSession], extras: List[str]) -> None:  # pragma: no cover
    """Ensure local dev-dependencies and calcipy extras are installed.

    See: https://github.com/cjolowicz/nox-poetry/issues/230#issuecomment-855445920

    """
    if read_package_name() == 'calcipy':
        session = cast(NPSession, session)
        session.poetry.installroot(extras=extras)
    else:
        session.install('.', f'calcipy[{",".join(extras)}]')

    if dev_deps := _installable_dev_dependencies():
        session.install(*dev_deps)


@nox_session(python=_get_pythons(), reuse_venv=True)
def tests(session: Union[NoxSession, NPSession]) -> None:  # pragma: no cover
    """Run doit test task for specified python versions."""
    _install_local(session, ['ddict', 'doc', 'lint', 'nox', 'stale', 'tags', 'test'])
    session.run(*shlex.split('pytest ./tests'), stdout=True)


@nox_session(python=_get_pythons()[-1:], reuse_venv=False)
def build_dist(session: Union[NoxSession, NPSession]) -> None:  # pragma: no cover
    """Build and test the project files within a controlled environment for repeatability."""
    dist_path = Path('dist')
    if_found_unlink(dist_path)

    # Support 'corallium' by re-implementing "session.poetry.build_package()", from:
    # https://github.com/cjolowicz/nox-poetry/blob/5772b66ebff8d5a3351a08ed402d3d31e48be5f8/src/nox_poetry/sessions.py#L233-L255
    # https://github.com/cjolowicz/nox-poetry/blob/5772b66ebff8d5a3351a08ed402d3d31e48be5f8/src/nox_poetry/poetry.py#L111-L154
    output = session.run(*shlex.split('poetry build --format=wheel --no-ansi'),
                         external=True, silent=True, stderr=None)
    output = cast(str, output)
    wheel = dist_path / output.split()[-1]
    path_wheel = wheel.resolve().as_uri()

    logger.text('Created wheel', path_wheel=path_wheel)
    # Install the wheel and check that imports without any of the optional dependencies
    session.install(path_wheel)
    session.run(*shlex.split('python scripts/check_imports.py'), stdout=True)


@nox_poetry_session(python=_get_pythons()[-1:], reuse_venv=True)
def build_check(session: NPSession) -> None:  # pragma: no cover
    """Check that the built output meets all checks."""
    # Build sdist and fix return URI, which will have file://...#egg=calcipy
    sdist_uri = session.poetry.build_package(distribution_format=DistributionFormat.SDIST)
    path_sdist = Path(url2pathname(urlparse(sdist_uri).path))
    logger.text_debug('Fixed sdist URI', sdist_uri=sdist_uri, path_sdist=path_sdist)
    # Check with pyroma
    session.install('pyroma>=4.0', '--upgrade')
    # required for "poetry.core.masonry.api" build backend
    session.run('python', '-m', 'pip', 'install', 'poetry>=1.3', stdout=True)
    session.run('pyroma', '--file', path_sdist.as_posix(), '--min=9', stdout=True)



"""Markdown Machine."""

import json
import re
from pathlib import Path

import pandas as pd
from beartype import beartype
from beartype.typing import Any, Callable, Dict, List, Optional
from corallium.file_helpers import read_lines
from corallium.log import logger
from transitions import Machine

from ..file_search import find_project_files_by_suffix
from ..invoke_helpers import get_project_path

HandlerLookupT = Dict[str, Callable[[str, Path], List[str]]]
"""Handler Lookup."""


class _ParseSkipError(RuntimeError):
    """Exception caught if the handler does not want to replace the text."""


class _ReplacementMachine(Machine):
    """State machine to replace content with user-specified handlers.

    Uses `{cts}` and `{cte}` to demarcate sections (short for calcipy_template start|end)

    """

    def __init__(self) -> None:
        """Initialize the state machine."""
        self.state_user = 'user'
        self.state_auto = 'autoformatted'
        transitions = [
            {
                'trigger': 'start_auto',
                'source': self.state_user,
                'dest': self.state_auto,
            }, {
                'trigger': 'end',
                'source': self.state_auto,
                'dest': self.state_user,
            },
        ]
        super().__init__(
            states=[self.state_user, self.state_auto],
            initial=self.state_user,
            transitions=transitions,
        )

    @beartype
    def parse(
        self, lines: List[str], handler_lookup: HandlerLookupT,
        path_file: Path,
    ) -> List[str]:
        """Parse lines and insert new_text based on provided handler_lookup.

        Args:
            lines: list of string from source file
            handler_lookup: Lookup dictionary for autoformatted sections
            path_file: optional path to the file. Only useful for debugging

        Returns:
            List[str]: modified list of strings

        """
        updated_lines = []
        for line in lines:
            updated_lines.extend(self._parse_line(line, handler_lookup, path_file))
        return updated_lines

    @beartype
    def _parse_line(
        self, line: str, handler_lookup: HandlerLookupT, path_file: Path,
    ) -> List[str]:
        """Parse lines and insert new_text based on provided handler_lookup.

        Args:
            line: single line
            handler_lookup: Lookup dictionary for autoformatted sections
            path_file: optional path to the file. Only useful for debugging

        Returns:
            List[str]: modified list of strings

        """
        lines: List[str] = []
        if '{cte}' in line and self.state == self.state_auto:  # end
            self.end()
        elif '{cts}' in line:  # start
            self.start_auto()
            matches = [text_match for text_match in handler_lookup if text_match in line]
            if len(matches) == 1:
                try:
                    lines.extend(handler_lookup[matches[0]](line, path_file))
                except _ParseSkipError:
                    lines.append(line)
                    self.end()
            else:
                logger.warning('Could not parse. Skipping:', line=line)
                lines.append(line)
                self.end()
        elif self.state == self.state_user:
            lines.append(line)
        # else: discard the lines in the auto-section
        return lines


_VAR_COMMENT_HTML = r'<!-- {cts} (?P<key>[^=]+)=(?P<value>[^;]+);'
"""Regex for extracting the variable from an HTML code comment."""


@beartype
def _parse_var_comment(line: str, matcher: str = _VAR_COMMENT_HTML) -> Dict[str, str]:
    """Parse the variable from a matching comment.

    Args:
        line: string from source file
        matcher: string regex pattern to match. Default is `_RE_VAR_COMMENT_HTML`

    Returns:
        Dict[str, str]: single key and value pair based on the parsed comment

    """
    if match := re.compile(matcher).match(line.strip()):
        matches = match.groupdict()
        return {matches['key']: matches['value']}
    return {}


@beartype
def _handle_source_file(line: str, path_file: Path) -> List[str]:
    """Replace commented sections in README with linked file contents.

    Args:
        line: first line of the section
        path_file: path to the file that contained the string

    Returns:
        List[str]: list of auto-formatted text

    """
    key, path_rel = [*_parse_var_comment(line).items()][0]
    path_base = get_project_path() if path_rel.startswith('/') else path_file.resolve().parent
    path_source = path_base / path_rel.lstrip('/')
    language = path_source.suffix.lstrip('.')
    lines_source = [f'```{language}', *read_lines(path_source), '```']
    if not path_source.is_file():
        logger.warning('Could not locate source file', path_source=path_source)

    line_start = f'<!-- {{cts}} {key}={path_rel}; -->'
    line_end = '<!-- {cte} -->'
    return [line_start, *lines_source, line_end]


@beartype
def _format_cov_table(coverage_data: Dict[str, Any]) -> List[str]:
    """Format code coverage data table as markdown.

    Args:
        coverage_data: dictionary created by `python -m coverage json`

    Returns:
        List[str]: list of string lines to insert

    """
    col_key_map = {
        'Statements': 'num_statements',
        'Missing': 'missing_lines',
        'Excluded': 'excluded_lines',
        'Coverage': 'percent_covered',
    }
    records = [
        {
            'File': f'`{Path(path_file).as_posix()}`',
            **{col: file_obj['summary'][key] for col, key in col_key_map.items()},
        } for path_file, file_obj in coverage_data['files'].items()
    ]
    records.append({
        'File': '**Totals**',
        **{col: coverage_data['totals'][key] for col, key in col_key_map.items()},
    })
    # Format table for Github Markdown
    df_cov = pd.DataFrame(records)
    df_cov['Coverage'] = df_cov['Coverage'].round(1).astype(str) + '%'
    lines_table = str(df_cov.to_markdown(index=False, tablefmt='github')).split('\n')
    short_date = coverage_data['meta']['timestamp'].split('T')[0]
    lines_table.extend(['', f'Generated on: {short_date}'])
    return lines_table


@beartype
def _handle_coverage(line: str, _path_file: Path, path_coverage: Optional[Path] = None) -> List[str]:
    """Read the coverage.json file and write a Markdown table to the README file.

    Args:
        line: first line of the section
        path_file: path to the file that contained the string

    Returns:
        List[str]: list of auto-formatted text

    Raises:
        _ParseSkipError: if the "coverage.json" file is not available

    """
    path_coverage = path_coverage or get_project_path() / 'coverage.json'
    if not path_coverage.is_file():
        msg = f'Could not locate: {path_coverage}'
        raise _ParseSkipError(msg)
    coverage_data = json.loads(path_coverage.read_text())
    lines_cov = _format_cov_table(coverage_data)
    line_end = '<!-- {cte} -->'
    return [line, *lines_cov, line_end]


@beartype
def write_autoformatted_md_sections(
    handler_lookup: Optional[HandlerLookupT] = None,
    paths_md: Optional[List[Path]] = None,
) -> None:
    """Populate the auto-formatted sections of markdown files with user-configured logic."""
    _lookup: HandlerLookupT = handler_lookup or {
        'COVERAGE ': _handle_coverage,
        'SOURCE_FILE=': _handle_source_file,
    }

    paths = paths_md or find_project_files_by_suffix(get_project_path()).get('md') or []
    for path_md in paths:
        logger.text_debug('Processing', path_md=path_md)
        if md_lines := _ReplacementMachine().parse(read_lines(path_md), _lookup, path_md):
            path_md.write_text('\n'.join(md_lines) + '\n', encoding='utf-8')



"""Lint CLI."""

from contextlib import suppress

from beartype import beartype
from beartype.typing import Optional
from corallium.file_helpers import read_package_name
from corallium.log import logger
from invoke.context import Context

from ..cli import task
from ..invoke_helpers import run
from .executable_utils import PRE_COMMIT_MESSAGE, check_installed, python_dir, python_m

# ==============================================================================
# Linting


@beartype
def _inner_task(
    ctx: Context,
    *,
    command: str,
    cli_args: str = '',
    run_as_module: bool = True,
    target: Optional[str] = None,
) -> None:
    """Shared task logic."""
    file_args = []
    with suppress(AttributeError):
        file_args = ctx.config.gto.file_args
    if file_args:
        target = ' '.join([str(_a) for _a in file_args])
    elif target is None:
        target = f'./{read_package_name()} ./tests'

    cmd = f'{python_m()} {command}' if run_as_module else f'{python_dir()}/{command}'
    run(ctx, f'{cmd} {target} {cli_args}'.strip())


@task(default=True)
def check(ctx: Context) -> None:
    """Run ruff as check-only."""
    _inner_task(ctx, command='ruff check')


@task()
def autopep8(ctx: Context) -> None:
    """Run autopep8.

    FYI: This is temporary until ruff implements white space rules
    https://github.com/charliermarsh/ruff/issues/970

    """
    cli_args = '--aggressive --recursive --in-place --max-line-length=120'
    _inner_task(ctx, command='autopep8', cli_args=cli_args)


@task(pre=[autopep8])
def fix(ctx: Context) -> None:
    """Run ruff and apply fixes."""
    _inner_task(ctx, command='ruff check', cli_args='--fix')


@task()
def watch(ctx: Context) -> None:
    """Run ruff as check-only."""
    _inner_task(ctx, command='ruff check', cli_args='--watch --show-source')


@task()
def flake8(ctx: Context) -> None:
    """Run flake8."""
    _inner_task(ctx, command='flake8', run_as_module=False)


@task(
    help={
        'report': 'if provided, show the pylint summary report',
    },
)
def pylint(ctx: Context, *, report: bool = False) -> None:
    """Run pylint."""
    cli_args = '--report=y' if report else ''
    _inner_task(ctx, command='pylint', cli_args=cli_args)


# ==============================================================================
# Security


@task()
def security(ctx: Context) -> None:
    """Attempt to identify possible security vulnerabilities."""
    logger.text('Note: Selectively override bandit with "# nosec"', is_header=True)
    pkg_name = read_package_name()
    run(ctx, f'{python_dir()}/bandit --recursive {pkg_name} -s B101')

    # See additional semgrep rules at:
    #   https://semgrep.dev/explore
    #   https://github.com/returntocorp/semgrep-rules/tree/develop/python
    #   https://awesomeopensource.com/project/returntocorp/semgrep-rules?categorypage=45
    semgrep_configs = ' '.join([  # noqa: FLY002
        '--config=p/ci',
        '--config=p/default',
        '--config=p/security-audit',
        '--config=r/bash',
        '--config=r/contrib',
        '--config=r/fingerprints',
        '--config=r/generic',
        '--config=r/json',
        '--config=r/python',
        '--config=r/terraform',
        '--config=r/yaml',
        '--exclude-rule=yaml.github-actions.security.third-party-action-not-pinned-to-commit-sha.third-party-action-not-pinned-to-commit-sha',
    ])
    logger.text('Note: Selectively override semgrep with "# nosem"', is_header=True)
    run(ctx, f'{python_dir()}/semgrep ci --autofix {semgrep_configs}')


# ==============================================================================
# Pre-Commit


@task(
    help={
        'no_update': 'Skip updating the pre-commit hooks',
    },
)
def pre_commit(ctx: Context, *, no_update: bool = False) -> None:
    """Run pre-commit."""
    check_installed(ctx, executable='pre-commit', message=PRE_COMMIT_MESSAGE)

    run(ctx, 'pre-commit install')
    if not no_update:
        run(ctx, 'pre-commit autoupdate')

    all_hook_stages = [
        'commit', 'merge-commit', 'push', 'prepare-commit-msg', 'commit-msg', 'post-checkout',
        'post-commit', 'post-merge', 'post-rewrite', 'manual',
    ]
    stages_cli = ' '.join(f'--hook-stage {stg}' for stg in all_hook_stages)
    run(ctx, f'pre-commit run --all-files {stages_cli}')



"""Tasks can be imported piecemeal or imported in their entirety from here."""

from beartype import beartype
from beartype.typing import Any, List, Union
from corallium.log import logger
from invoke.context import Context
from invoke.tasks import Call

from ..cli import task
from . import cl, doc, lint, nox, pack, stale, tags, test, types
from ._invoke import Collection, DeferedTask, _build_task
from .defaults import new_collection

# "ns" will be recognized by Collection.from_module(all_tasks)
# https://docs.pyinvoke.org/en/stable/api/collection.html#invoke.collection.Collection.from_module
ns = new_collection()
ns.add_collection(Collection.from_module(cl))
ns.add_collection(Collection.from_module(doc))
ns.add_collection(Collection.from_module(lint))
ns.add_collection(Collection.from_module(nox))
ns.add_collection(Collection.from_module(pack))
ns.add_collection(Collection.from_module(stale))
ns.add_collection(Collection.from_module(tags))
ns.add_collection(Collection.from_module(test))
ns.add_collection(Collection.from_module(types))


@task(
    help={
        'message': 'String message to display',
    },
    show_task_info=False,
)
def summary(_ctx: Context, *, message: str) -> None:
    """Summary Task."""
    logger.text(message, is_header=True)


@task(
    help={
        'index': 'Current index (0-indexed)',
        'total': 'Total steps',
    },
    show_task_info=False,
)
def progress(_ctx: Context, *, index: int, total: int) -> None:
    """Progress Task."""
    logger.text('Progress', is_header=True, index=index + 1, total=total)


TaskList = List[Union[Call, DeferedTask]]
"""List of wrapped or normal task functions."""


@beartype
def with_progress(
    items: Any,  # PLANNED: TaskList,
    offset: int = 0,
) -> TaskList:
    """Inject intermediary 'progress' tasks.

    Args:
        items: list of tasks
        offset: Optional offset to shift counters

    """
    task_items = [_build_task(_t) for _t in items]
    message = 'Running tasks: ' + ', '.join([str(_t.__name__) for _t in task_items])
    # FIXME: tasks: TaskList = [summary.with_kwargs(message=message)]
    tasks = [summary.with_kwargs(message=message)]

    total = len(task_items) + offset
    for idx, item in enumerate(task_items):
        # pyright: ignore[reportGeneralTypeIssues]
        tasks.extend([progress.with_kwargs(index=idx + offset, total=total), item])
    return tasks  # pyright: ignore[reportGeneralTypeIssues]


_MAIN_TASKS = [
    lint.fix,
    types.mypy,
    types.pyright,
    nox.noxfile.with_kwargs(session='tests'),  # pyright: ignore[reportGeneralTypeIssues]
    lint.pre_commit.with_kwargs(no_update=True),  # pyright: ignore[reportGeneralTypeIssues]
    lint.security,
    tags.collect_code_tags,
    cl.write,
    pack.lock,
    test.coverage,
    doc.build,
    stale.check_for_stale_packages,
]
_OTHER_TASKS = [
    test.check,
    lint.flake8,
    lint.pylint,
    pack.check_licenses,
    test.step,
]


@task(post=with_progress(_MAIN_TASKS))
def main(_ctx: Context) -> None:
    """Main task pipeline."""


@task(post=with_progress(_OTHER_TASKS))  # pyright: ignore[reportGeneralTypeIssues]
def other(_ctx: Context) -> None:
    """Run tasks that are otherwise not exercised in main."""


@task(
    help=cl.bump.help,
    post=with_progress(
        [  # pyright: ignore[reportGeneralTypeIssues]
            pack.lock,
            doc.build,
            doc.deploy,
            pack.publish,
        ],
        offset=1,
    ),
)
def release(ctx: Context, *, suffix: cl.SuffixT = None) -> None:
    """Release pipeline."""
    cl.bumpz(ctx, suffix=suffix)


ns.add_task(main)
ns.add_task(other)
ns.add_task(release)



"""Find Files."""

from collections import defaultdict
from pathlib import Path

from beartype import beartype
from beartype.typing import Dict, List, Optional
from corallium.log import logger
from corallium.shell import capture_shell


@beartype
def _zsplit(stdout: str) -> List[str]:
    """Split output from git when used with `-z`."""
    return [item for item in stdout.split('\0') if item]


@beartype
def _get_all_files(*, cwd: Path) -> List[str]:
    """Get all files using git. Modified `pre_commit.git.get_all_files` to accept `cwd`.

    https://github.com/pre-commit/pre-commit/blob/488b1999f36cac62b6b0d9bc8eae99418ae5c226/pre_commit/git.py#L153

    Args:
        cwd: current working directory to pass to `subprocess.Popen`

    Returns:
        List[str]: list of all file paths relative to the `cwd`

    """
    return _zsplit(capture_shell('git ls-files -z', cwd=cwd))


@beartype
def _filter_files(rel_filepaths: List[str], ignore_patterns: List[str]) -> List[str]:
    """Filter a list of string file paths with specified ignore patterns in glob syntax.

    Args:
        rel_filepaths: list of string file paths
        ignore_patterns: glob ignore patterns

    Returns:
        List[str]: list of all non-ignored file path names

    """
    if ignore_patterns:
        matches = []
        for _fp in rel_filepaths:
            pth = Path(_fp).resolve()
            if not any(pth.match(pat) for pat in ignore_patterns):
                matches.append(_fp)
        return matches
    return rel_filepaths


@beartype
def find_project_files(path_project: Path, ignore_patterns: List[str]) -> List[Path]:
    """Find project files in git version control.

    > Note: uses the relative project directory and verifies that each file exists

    Args:
        path_project: Path to the project directory
        ignore_patterns: glob ignore patterns

    Returns:
        Dict[str, List[Path]]: where keys are the suffix (without leading dot) and values the list of paths

    """
    file_paths = []
    rel_filepaths = _get_all_files(cwd=path_project)
    filtered_rel_files = _filter_files(rel_filepaths=rel_filepaths, ignore_patterns=ignore_patterns)
    for rel_file in filtered_rel_files:
        path_file = path_project / rel_file
        if path_file.is_file():
            file_paths.append(path_file)
        else:  # pragma: no cover
            logger.warning('Could not find the specified file', path_file=path_file)
    return file_paths


# TODO: Consider adding a configuration item for ignore_patterns
@beartype
def find_project_files_by_suffix(
    path_project: Path, *, ignore_patterns: Optional[List[str]] = None,
) -> Dict[str, List[Path]]:
    """Find project files in git version control.

    > Note: uses the relative project directory and verifies that each file exists

    Args:
        path_project: Path to the project directory
        ignore_patterns: glob ignore patterns

    Returns:
        Dict[str, List[Path]]: where keys are the suffix (without leading dot) and values the list of paths

    """
    file_lookup: Dict[str, List[Path]] = defaultdict(list)
    for path_file in find_project_files(path_project, ignore_patterns or []):
        file_lookup[path_file.suffix.lstrip('.')].append(path_file)
    return dict(file_lookup)



"""Utilities for working in calcipy's python environment."""

import sys
from contextlib import suppress
from functools import lru_cache
from pathlib import Path

from beartype import beartype
from invoke.context import Context

from ..invoke_helpers import run


@lru_cache(maxsize=1)
@beartype
def resolve_python() -> Path:
    """Resolve the user's Python path based on `sys`."""
    python_path = Path(sys.executable)
    with suppress(ValueError):
        return python_path.relative_to(Path.cwd())
    return python_path


@lru_cache(maxsize=1)
@beartype
def python_dir() -> str:
    """Runs an executable from the currently active Python directory."""
    return str(resolve_python().parent)


@lru_cache(maxsize=1)
@beartype
def python_m() -> str:
    """Return the active python path and `-m` flag."""
    return f'{resolve_python()} -m'


PYRIGHT_MESSAGE = """
`pyright` was not found and must be installed separately (such as 'brew install pyright' on Mac).
    See the online documentation for your system: https://microsoft.github.io/pyright/#/installation
"""
PRE_COMMIT_MESSAGE = """
`pre-commit` was not found and must be installed separately (such as 'brew install pre-commit' on Mac).
    See the online documentation for your system: https://pre-commit.com/#install
"""
GH_MESSAGE = """
`gh` was not found and must be installed separately (such as 'brew install gh' on Mac).
    See the online documentation for your system: https://cli.github.com/
"""


@beartype
def check_installed(ctx: Context, executable: str, message: str) -> None:
    """If the required executable isn't present, raise a clear user error."""
    res = run(ctx, f'which {executable}', warn=True, hide=True)
    if not res or res.exited == 1:
        raise RuntimeError(message)



"""Test CLI."""

from pathlib import Path

from beartype import beartype
from beartype.typing import Optional
from corallium.file_helpers import open_in_browser, read_package_name
from invoke.context import Context

from ..cli import task
from ..experiments import check_duplicate_test_names
from ..invoke_helpers import run
from .defaults import from_ctx
from .executable_utils import python_dir

_STEPWISE_ARGS = ' --failed-first --new-first --exitfirst -vv --no-cov'


@beartype
def _inner_task(
    ctx: Context,
    *,
    command: str = 'pytest',
    cli_args: str = '',
    keyword: str = '',
    marker: str = '',
    min_cover: int = 0,
) -> None:
    """Shared task logic."""
    if keyword:
        cli_args += f' -k "{keyword}"'
    if marker:
        cli_args += f' -m "{marker}"'
    if fail_under := min_cover or int(from_ctx(ctx, 'test', 'min_cover')):
        cli_args += f' --cov-fail-under={fail_under}'
    run(ctx, f'{python_dir()}/{command} ./tests{cli_args}')


@task()
def check(_ctx: Context) -> None:
    """Run pytest checks, such as identifying ."""
    if duplciates := check_duplicate_test_names.run(Path('tests')):
        raise RuntimeError(f'Duplicate test names found ({duplciates}). See above for details.')  # noqa: EM102


KM_HELP = {
    # See: https://docs.pytest.org/en/latest/usage.html#specifying-tests-selecting-tests
    'keyword': 'Only run tests that match the string pattern',
    'marker': 'Only run tests matching given mark expression',
}


@task(
    default=True,
    help={
        'min_cover': 'Fail if coverage less than threshold',
        **KM_HELP,
    },
)
def pytest(ctx: Context, *, keyword: str = '', marker: str = '', min_cover: int = 0) -> None:
    """Run pytest with default arguments."""
    pkg_name = read_package_name()
    durations = '--durations=25 --durations-min="0.1"'
    _inner_task(ctx,
                cli_args=f' --cov={pkg_name} --cov-branch --cov-report=term-missing {durations}',
                keyword=keyword, marker=marker, min_cover=min_cover)


@task(help=KM_HELP)
def step(ctx: Context, *, keyword: str = '', marker: str = '') -> None:
    """Run pytest optimized to stop on first error."""
    _inner_task(ctx, cli_args=_STEPWISE_ARGS, keyword=keyword, marker=marker)


@task(help=KM_HELP)
def watch(ctx: Context, *, keyword: str = '', marker: str = '') -> None:
    """Run pytest with polling and optimized to stop on first error."""
    _inner_task(ctx, cli_args=_STEPWISE_ARGS, keyword=keyword, marker=marker, command='ptw . --now')


@task(
    help={
        'min_cover': 'Fail if coverage less than threshold',
        'out_dir': 'Optional path to coverage directory. Typically ".cover" or "releases/tests"',
        'view': 'If True, open the created files',
    },
)
def coverage(ctx: Context, *, min_cover: int = 0, out_dir: Optional[str] = None, view: bool = False) -> None:
    """Generate useful coverage outputs after running pytest.

    Creates `coverage.json` used in `doc.build`

    """
    pkg_name = read_package_name()
    _inner_task(ctx, cli_args='', min_cover=min_cover,
                command=f'coverage run --branch --source={pkg_name} --module pytest')

    cov_dir = Path(out_dir or from_ctx(ctx, 'test', 'out_dir'))
    cov_dir.mkdir(exist_ok=True, parents=True)
    print('')  # noqa: T201
    for cli_args in (
        'report --show-missing',  # Write to STDOUT
        f'html --directory={cov_dir}',  # Write to HTML
        'json',  # Create coverage.json file for "_handle_coverage"
    ):
        run(ctx, f'{python_dir()}/coverage {cli_args}')

    if view:  # pragma: no cover
        open_in_browser(cov_dir / 'index.html')



"""Packaging CLI."""

from corallium import file_helpers  # Required for mocking read_pyproject
from corallium.file_helpers import LOCK, PROJECT_TOML
from corallium.log import logger
from invoke.context import Context

from .. import can_skip  # Required for mocking can_skip.can_skip
from ..cli import task
from ..experiments import bump_programmatically
from ..invoke_helpers import run
from .executable_utils import python_dir


@task()
def install_extras(ctx: Context) -> None:
    """Run poetry install with all extras."""
    poetry_config = file_helpers.read_pyproject()['tool']['poetry']
    extras = (poetry_config.get('extras') or {}).keys()
    run(ctx, ' '.join(['poetry install --sync', *[f'--extras={ex}' for ex in extras]]))


@task()
def lock(ctx: Context) -> None:
    """Ensure poetry.lock is  up-to-date."""
    if can_skip.can_skip(prerequisites=[PROJECT_TOML], targets=[LOCK]):
        return  # Exit early

    run(ctx, 'poetry lock --no-update')


@task(
    help={
        'to_test_pypi': 'Publish to the TestPyPi repository',
    },
)
def publish(ctx: Context, *, to_test_pypi: bool = False) -> None:
    """Build the distributed format(s) and publish."""
    run(ctx, f'{python_dir()}/nox --error-on-missing-interpreters --session build_dist build_check')

    cmd = 'poetry publish'
    if to_test_pypi:
        cmd += ' --repository testpypi'
    run(ctx, cmd)


@task()
def check_licenses(ctx: Context) -> None:
    """Check licenses for compatibility with `licensecheck`."""
    res = run(ctx, 'which licensecheck', warn=True, hide=True)
    if not res or res.exited == 1:
        logger.warning('`licensecheck` not found. installing with pipx')
        run(ctx, 'pipx install licensecheck')
    run(ctx, 'licensecheck')


@task(
    help={
        'tag': 'Last tag, can be provided with `--tag="$(git tag -l "v*" | sort | tail -n 1)"`',
        'tag_prefix': 'Optional tag prefix, such as "v"',
        'pkg_name': 'Optional package name. If not provided, will read the poetry pyproject.toml file',
    },
)
def bump_tag(ctx: Context, *, tag: str, tag_prefix: str = '', pkg_name: str = '') -> None:  # noqa: ARG001
    """Experiment with bumping the git tag using `griffe`.

    Example for `calcipy`:

    ```sh
    ./run pack.bump-tag --tag="$(git tag -l "*" | sort | head -n 5 | tail -n 1)" --tag-prefix=""
    ```

    """
    if not tag:
        raise ValueError('tag must not be empty')
    if not pkg_name:
        poetry_config = file_helpers.read_pyproject()['tool']['poetry']
        pkg_name = poetry_config['name']

    new_version = bump_programmatically.bump_tag(
        pkg_name=pkg_name,
        tag=tag,
        tag_prefix=tag_prefix,
    )
    logger.text(new_version)



"""Start the command line program."""

from types import ModuleType

from beartype import beartype
from beartype.typing import List

from . import __pkg_name__, __version__
from .cli import start_program
from .tasks._invoke import Collection


@beartype
def start() -> None:  # pragma: no cover
    """Run the customized Invoke Program."""
    from .tasks import all_tasks
    start_program(__pkg_name__, __version__, all_tasks)


@beartype
def _start_subset(modules: List[ModuleType]) -> None:  # pragma: no cover
    """Run the specified subset."""
    from .tasks.defaults import new_collection

    ns = new_collection()
    for module in modules:
        ns.add_collection(Collection.from_module(module))

    start_program(__pkg_name__, __version__, collection=ns)


@beartype
def start_lint() -> None:  # pragma: no cover
    """Run CLI with only the lint namespace."""
    from .tasks import lint
    _start_subset([lint])


@beartype
def start_pack() -> None:  # pragma: no cover
    """Run CLI with only the pack namespace."""
    from .tasks import pack
    _start_subset([pack])


@beartype
def start_tags() -> None:  # pragma: no cover
    """Run CLI with only the tags namespace."""
    from .tasks import tags
    _start_subset([tags])


@beartype
def start_types() -> None:  # pragma: no cover
    """Run CLI with only the types namespace."""
    from .tasks import types
    _start_subset([types])



"""Types CLI."""

from beartype import beartype
from corallium.file_helpers import read_package_name
from invoke.context import Context

from ..cli import task
from ..invoke_helpers import run
from .executable_utils import PYRIGHT_MESSAGE, check_installed, python_dir


@beartype
def _inner_task(ctx: Context, *, command: str, cli_args: str = '') -> None:
    """Shared task logic."""
    pkg_name = read_package_name()
    run(ctx, f'{command} {pkg_name} {cli_args}'.strip())


@task()
def pyright(ctx: Context) -> None:
    """Run pyright."""
    check_installed(ctx, executable='pyright', message=PYRIGHT_MESSAGE)
    _inner_task(ctx, command='pyright')


@task()
def mypy(ctx: Context) -> None:
    """Run mypy."""
    _inner_task(ctx, command=f'{python_dir()}/mypy')



"""Experiment with checking for duplicate test names."""

import ast
from pathlib import Path

from beartype import beartype
from beartype.typing import List, Union
from corallium.log import logger


@beartype
def _show_info(function: Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> None:
    """Print info about the function."""
    logger.info('> name', name=function.name)
    if function.args.args:
        logger.info('\t args', args=function.args.args)


@beartype
def run(test_path: Path) -> List[str]:  # noqa: C901,CAC001  # pylint: disable=too-complex
    """Check for duplicates in the test suite.

    Inspired by: https://stackoverflow.com/a/67840804/3219667

    """
    summary = set()
    duplicates = []

    for path_test in test_path.rglob('test_*.py'):  # pylint: disable=too-many-nested-blocks
        logger.info(path_test.as_posix())
        parsed_ast = ast.parse(path_test.read_text())

        for node in parsed_ast.body:
            if isinstance(node, ast.FunctionDef):
                if node.name in summary and node.name.startswith('test_'):
                    duplicates.append(node.name)
                summary.add(node.name)
                _show_info(node)
            elif isinstance(node, ast.ClassDef):
                logger.info('Class name', name=node.name)
                for method in node.body:
                    if isinstance(method, ast.FunctionDef):
                        _show_info(method)

        for node in ast.walk(parsed_ast):  # type: ignore[assignment]
            if (
                isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
                and node.name not in summary
            ):
                logger.info('Found new function(s) through walking')
                _show_info(node)
                summary.add(node.name)

    if duplicates:
        logger.error('Found Duplicates', duplicates=duplicates)
    return duplicates


if __name__ == '__main__':  # pragma: no cover
    run(Path('tests'))



"""Changelog CLI."""

from beartype import beartype
from beartype.typing import Literal, Optional
from invoke.context import Context

from ..cli import task
from ..invoke_helpers import get_doc_subdir, get_project_path, run
from .executable_utils import GH_MESSAGE, check_installed, python_dir

SuffixT = Optional[Literal['alpha', 'beta', 'rc']]
"""Prerelease Suffix Type."""


@task()
def write(ctx: Context) -> None:
    """Write a Changelog file with the raw Git history.

    Resources:

    - https://keepachangelog.com/en/1.0.0/
    - https://www.conventionalcommits.org/en/v1.0.0/
    - https://writingfordevelopers.substack.com/p/how-to-write-a-commit-message
    - https://chris.beams.io/posts/git-commit/
    - https://semver.org/
    - https://calver.org/

    Returns:
        List[DoitAction]: doit actions

    """
    run(ctx, f'{python_dir()}/cz changelog')
    path_cl = get_project_path() / 'CHANGELOG.md'
    if not path_cl.is_file():
        msg = f'Could not locate the changelog at: {path_cl}'
        raise FileNotFoundError(msg)
    path_cl.replace(get_doc_subdir() / path_cl.name)


@beartype
def bumpz(ctx: Context, *, suffix: SuffixT = None) -> None:
    """Bumps project version based on commits & settings in pyproject.toml."""
    check_installed(ctx, executable='gh', message=GH_MESSAGE)

    opt_cz_args = f' --prerelease={suffix}' if suffix else ''
    run(ctx, f'{python_dir()}/cz bump{opt_cz_args} --annotated-tag --no-verify --gpg-sign')

    run(ctx, 'git push origin --tags --no-verify')

    get_last_tag = 'git tag --list --sort=-creatordate | head -n 1'
    opt_gh_args = ' --prerelease' if suffix else ''
    run(ctx, f'gh release create --generate-notes $({get_last_tag}){opt_gh_args}')


@task(
    pre=[write],
    help={
        'suffix': 'Specify prerelease suffix for version bump (alpha, beta, rc)',
    },
)
def bump(ctx: Context, *, suffix: SuffixT = None) -> None:
    """Bumps project version based on commits & settings in pyproject.toml."""
    bumpz(ctx, suffix=suffix)
